"""Constants for the Mazda Connected Services integration."""

DOMAIN = "mazda"

DATA_CLIENT = "mazda_client"
DATA_COORDINATOR = "coordinator"

MAZDA_REGIONS = {"MNAO": "North America", "MME": "Europe", "MJO": "Japan"}
