"""Define ReCollect Waste constants."""
import logging

DOMAIN = "recollect_waste"

LOGGER = logging.getLogger(__package__)

CONF_PLACE_ID = "place_id"
CONF_SERVICE_ID = "service_id"

DATA_COORDINATOR = "coordinator"
